"""For convenience, this module exposes main classes and functions of the library"""
from jesth.document import Document
from jesth.section import Section
from jesth.parser import read, parse, Parser
from jesth.renderer import render, write
from jesth.converter import make_dict, flatten_dict, ValueConverter
from jesth.misc import split_key_value


__all__ = ["Document", "Section", "parse", "render", "read", "write",
           "make_dict", "flatten_dict", "ValueConverter", "split_key_value",
           "Parser"]



"""Boxes to hold contents like Jesth Comments, raw texts, or whitespaces..."""
import threading


class RawString(str):
    """Box to hold raw string, i.e., string between two single quotes"""
    pass


class Text(str):
    """Box to hold Text, i.e., multiline string with unicode escape sequences support"""
    pass


class RawText(str):
    """Box to hold raw Text, i.e., multiline string without unicode escape sequences support"""
    pass


class HexInt(int):
    """Box to hold hexadecimal integer"""
    pass


class OctInt(int):
    """Box to hold octal integer"""
    pass


class BinInt(int):
    """Box to hold binary integer"""
    pass


class CommentID(int):
    """Box to hold an unique comment id"""
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, N.generate_id())


class Comment(str):
    """Box to hold a Jesth comment"""
    pass


class WhitespaceID(int):
    """Box to hold an unique whitespace id"""
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, N.generate_id())


class Whitespace(str):
    """Box to hold a whitespace"""
    pass


class N:
    """Private class to generate unique identifiers (thread-safe)"""
    _lock = threading.Lock()
    _count = 0

    @classmethod
    def generate_id(cls):
        """Generate an unique integer identifier (thread-safe)"""
        with cls._lock:
            cls._count += 1
            return cls._count



"""This module exposes render and write functions"""
import pathlib
from jesth import misc


def render(document, *headers, spacing=1):
    """
    Convert a jesth document into plain text

    [parameters]
    - document: an instance of jesth.document.Document
    - *headers: Headers of sections to render. Omitting this will render the entire section
    - spacing: number of empty lines between two adjacent sections. Defaults to 1 empty line.

    [return]
    Returns a string that contains sections (each made of square-brackets
    delimited header and associated body). This string can be stored in a file,
    thus creating a JesthFile
    """
    cache = list()
    for i, section in enumerate(document.sections):
        if headers and section.header not in headers:
            continue
        safe_body = [misc.add_leading_backslashes(line)
                     for line in section.body]
        body = "\n".join(safe_body)
        header = ("" if i == 0 and section.header == ""
                  else "[{}]".format(section.header))
        if body:
            header = header + "\n" if header else header
            text = "{header}{body}".format(header=header, body=body)
        else:
            text = "{header}".format(header=header)
        cache.append(text)
    spacing = ("\n" * spacing) + "\n"
    return spacing.join(cache)


def write(document, destination):
    """
    Convert a jesth document into plain text then write it in a file

    [parameters]
    - document: an instance of jesth.document.Document
    - destination: path string, or pathlib.Path instance to a file in which
     the rendered Jesth text can be saved.

    [return]
    Returns the rendered plain text
    """
    data = render(document)
    if isinstance(destination, pathlib.Path) or isinstance(destination, str):
        _write_to_file(destination, data)
    else:
        destination.write(data.encode("utf-8"))
    return data


def _write_to_file(destination, data):
    if isinstance(destination, pathlib.Path):
        destination = str(destination.resolve())
    misc.ensure_parent_dir(destination)
    with open(destination, "w", encoding="utf-8") as file:
        file.write(data)



import os.path


# general constants
HOME = os.path.expanduser("~")
PYRUSTIC_HOME = os.path.join(HOME, "pyrustic")
JESTH_DIR = os.path.join(PYRUSTIC_HOME, "jesth")  # free for storing random jesthfiles
INDENT_WIDTH = 4  # Four spaces



"""Classes and functions to convert a section body into a Python dict, or flatten a Python dict into a section body"""
import datetime
import base64
from decimal import Decimal
from collections import OrderedDict, namedtuple
from jesth import error, misc, const
from jesth.box import HexInt, OctInt, BinInt, RawString,\
    Text, RawText, CommentID, Comment, WhitespaceID, Whitespace


def make_dict(body, value_converter=None, strict=True):
    """
    Convert a section body into a Python dict

    [parameters]
    - body: string or list of strings
    - value_converter: instance of ValueConverter to customize value encoding
    - strict: when strict is True, comments and whitespaces aren't preserved

    [return]
    Returns a Python dict (OrderedDict by default, customizable with ValueConverter)

    [exceptions]
    - jesth.error.ConversionError: raised when an error occured while doing conversion
    """
    value_converter = value_converter if value_converter else ValueConverter()
    # base form (as a regular string)
    if isinstance(body, str):
        body = body.split("\n")
    return BaseToDict.run(body, value_converter, strict)


def flatten_dict(data, value_converter=None):
    """
    Convert a Python dict into a list of strings (section body in the base form)

    [parameters]
    - data: Python dict (I do recommend an OrderedDict)
    - value_converter: instance of ValueConverter to customize value encoding

    [return]
    Returns a list of strings
    """
    value_converter = value_converter if value_converter else ValueConverter()
    return DictToBase.run(data, value_converter)


class BaseToDict:
    @staticmethod
    def run(lines, value_converter=None, strict=True):
        value_converter = value_converter if value_converter else ValueConverter()
        stack = list()
        #root_container = OrderedDict()
        root_container = value_converter.dict_type()
        root_context = Context("dict", root_container, 0)
        stack.append(root_context)
        for lineno, line in enumerate(lines):
            # remove line breaks
            line = line.replace("\n", "\\n")
            context = stack[-1]
            line, indents = BaseToDict._prepare_line(line, context)
            # same indent
            if indents == context.indents:
                pass
            # lower indent
            elif indents < context.indents:
                BaseToDict._cleanup_stack(stack, value_converter, indents)
            # exaggerated forward indent
            else:
                msg = "Expected {} or less indents at line '{}'"
                msg = msg.format(context.indents, line)
                raise error.IndentError(msg)
            # process line
            try:
                BaseToDict._process_line(line, stack, value_converter,
                                         strict)
            except Exception as e:
                msg = "Error on line {}: {}{}"
                str_e = str(e)
                str_e = ": " + str_e if str_e else str_e
                msg = msg.format(lineno+1, type(e).__name__, str_e)
                raise type(e)(msg)
        # cleanup stack
        BaseToDict._cleanup_stack(stack, value_converter, 0)
        # end
        #return BaseToDict._finalize_root_dict(root_context, value_converter)
        return root_container

    @staticmethod
    def _prepare_line(line, context):
        indent_width = const.INDENT_WIDTH
        # empty or whitespace within Text context
        #if context.name in ("bin", "raw", "text"):
        if not line:
            return "", context.indents
        elif line.isspace():
            if len(line) <= context.indents * indent_width:
                return "", context.indents
            else:
                line = line[context.indents * indent_width:]
                return line, context.indents
        # unindent
        indents = misc.count_indents(line)
        line = line[indents * indent_width:]
        return line, indents

    @staticmethod
    def _process_line(line, stack, value_converter, strict):
        if not stack:
            return
        context = stack[-1]
        name = context.name
        collection = context.collection
        indents = context.indents
        if name == "bin":
            collection.append(line)
        elif name == "dict":
            # process comments
            if line.startswith("#"):
                if strict:
                    return
                comment_id = CommentID()
                collection[comment_id] = Comment(line)
            # process whitespaces
            elif not line or line.isspace():
                if strict:
                    return
                whitespace_id = WhitespaceID()
                collection[whitespace_id] = Whitespace()
            # process key value
            else:
                key, val = misc.split_key_value(line, sep="=", strip_whitespace=True)
                value = value_converter.decode(val)
                collection[key] = value
                container_name = BaseToDict._check_container_tag(val)
                if container_name:
                    BaseToDict._update_stack(stack, container_name, value, indents+1)
        elif name == "list":
            # process comments
            if line.startswith("#"):
                if strict:
                    return
                comment = Comment(line)
                collection.append(comment)
            # process whitespaces
            elif not line or line.isspace():
                if strict:
                    return
                whitespace = Whitespace(line)
                collection.append(whitespace)
            # process key value
            else:
                value = value_converter.decode(line)
                collection.append(value)
                container_name = BaseToDict._check_container_tag(line)
                if container_name:
                    BaseToDict._update_stack(stack, container_name, value, indents+1)
        elif name == "raw":
            collection.append(line)
        elif name == "text":
            collection.append(line)
        else:
            msg = "Unknown box '{}'. Expected bin, dict, list, text, or raw."
            msg = msg.format(name)
            raise error.Error(msg)

    @staticmethod
    def _check_container_tag(value):
        if not value.startswith("(") or not value.endswith(")"):
            return None
        value = value.strip("()")
        if value not in ("bin", "dict", "list", "raw", "text"):
            return None
        return value

    @staticmethod
    def _update_stack(stack, container_name,
                      container, indents):
        new_context = Context(container_name, container, indents)
        stack.append(new_context)

    @staticmethod
    def _cleanup_stack(stack, value_converter, indents):
        if not stack:
            return
        while True:
            if len(stack) == 1:
                break
            context = stack[-1]
            parent_context = stack[-2]
            if indents >= context.indents:
                break
            # bin
            if context.name == "bin":
                BaseToDict._finalize_bin(parent_context, context,
                                         value_converter)
            # dict
            #elif context.name == "dict":
            #    BaseToDict._finalize_dict(parent_context, context,
            #                              value_converter)
            # list
            #elif context.name == "list":
            #    BaseToDict._finalize_list(parent_context, context,
            #                              value_converter)
            # raw
            elif context.name == "raw":
                BaseToDict._finalize_raw(parent_context, context,
                                         value_converter)
            # text
            elif context.name == "text":
                BaseToDict._finalize_text(parent_context, context,
                                          value_converter)
            # delete the latest context
            del stack[-1]

    @staticmethod
    def _finalize_root_dict(root_context, value_converter):
        default_dict_type = value_converter.dict_type
        collection = root_context.collection
        if type(collection) is default_dict_type:
            return collection
        return default_dict_type(collection)

    @staticmethod
    def _finalize_bin(parent_context, context, value_converter):
        bin_block, whitespace_block = BaseToDict._split_block(context.collection)
        value = [item.strip() for item in bin_block if item]
        bin_data = value_converter.bin_decoder(value)
        #default_bin_type = value_converter.bin_type
        #if type(bin_data) is not default_bin_type:
        #    bin_data = default_bin_type(bin_data)
        BaseToDict._update_parent_context(parent_context, bin_data)
        for line in whitespace_block:
            BaseToDict._add_whitespace_to_parent_context(parent_context, line)
    """
    @staticmethod
    def _finalize_dict(parent_context, context, value_converter):
        default_dict_type = value_converter.dict_type
        box = context.box
        if type(box) is default_dict_type:
            return
        box = default_dict_type(box)
        BaseToDict._update_parent_context(parent_context, box)

    @staticmethod
    def _finalize_list(parent_context, context, value_converter):
        default_raw_type = value_converter.list_type
        box = context.box
        if type(box) is default_raw_type:
            return
        box = default_raw_type(box)
        BaseToDict._update_parent_context(parent_context, box)
    """

    @staticmethod
    def _finalize_raw(parent_context, context, value_converter):
        # concatenate text lines
        raw_block, whitespace_block = BaseToDict._split_block(context.collection)
        text = value_converter.raw_decoder(raw_block)

        #default_raw_type = value_converter.raw_type
        #if type(text) is not default_raw_type:
        #    text = default_raw_type(text)
        BaseToDict._update_parent_context(parent_context, text)
        for line in whitespace_block:
            BaseToDict._add_whitespace_to_parent_context(parent_context, line)
        return
        #default_raw_type = value_converter.raw_type
        #box = context.box
        #if type(box) is default_raw_type:
        #    return
        #box = default_raw_type(box)
        #BaseToDict._update_parent_context(parent_context, box)

    @staticmethod
    def _finalize_text(parent_context, context, value_converter):
        # concatenate text lines
        text_block, whitespace_block = BaseToDict._split_block(context.collection)
        text = value_converter.text_decoder(text_block)
        #default_text_type = value_converter.text_type
        #if type(text) is not default_text_type:
        #    text = default_text_type(text)
        BaseToDict._update_parent_context(parent_context, text)
        for line in whitespace_block:
            BaseToDict._add_whitespace_to_parent_context(parent_context, line)

    @staticmethod
    def _split_block(block):
        x, y = block, list()
        for i, line in enumerate(reversed(block)):
            if line.rstrip() == "---":
                n = len(block)
                a = n - (i + 1)
                b = a + 1
                x, y = block[0:a], block[b:]
                break
        return x, y

    @staticmethod
    def _update_parent_context(parent_context, data):
        if parent_context.name == "dict":
            key = list(parent_context.collection.keys())[-1]
            parent_context.collection[key] = data
        elif parent_context.name == "list":
            parent_context.collection[-1] = data
        else:
            msg = "Invalid parent name '{}'"
            msg = msg.format(parent_context.name)
            raise error.Error(msg)

    @staticmethod
    def _add_whitespace_to_parent_context(parent_context, data):
        whitespace = Whitespace(data)
        if parent_context.name == "dict":
            whitespace_id = WhitespaceID()
            parent_context.collection[whitespace_id] = whitespace
        elif parent_context.name == "list":
            parent_context.collection[-1] = whitespace
        else:
            msg = "Invalid parent name '{}'"
            msg = msg.format(parent_context.name)
            raise error.Error(msg)


class DictToBase:
    @staticmethod
    def run(data, value_converter=None):
        value_converter = value_converter if value_converter else ValueConverter()
        if type(data) not in value_converter.dict_types:
            msg = "Expected: {}".format(", ".join(value_converter.dict_types))
            raise error.Error(msg)
        return DictToBase._loop(data, value_converter)

    @staticmethod
    def _loop(data, value_converter, base=None, stack=None, cached_refs=None):
        base = base if base else list()
        cached_refs = cached_refs if cached_refs else list()
        cached_refs.append(data)
        if not stack:
            stack = list()
            #root_container = value_converter.dict_type()
            root_context = Context("dict", data, 0)
            stack.append(root_context)
        context = stack[-1]
        if context.name == "dict":
            DictToBase._handle_dict(context.collection, context,
                                    value_converter, base, stack,
                                    cached_refs)
        elif context.name == "list":
            DictToBase._handle_list(context.collection, context,
                                    value_converter, base, stack,
                                    cached_refs)
        return base

    @staticmethod
    def _handle_dict(data, context, value_converter, base, stack, cached_refs):
        indent_str = " " * const.INDENT_WIDTH
        context_indent_str = indent_str * context.indents
        for key, value in data.items():
            # comments
            if isinstance(key, CommentID):
                value = value.replace("\n", "\\n")
                base.append(context_indent_str + value)
            # whitespace
            elif isinstance(key, WhitespaceID):
                #base.append(context_indent_str + value)
                base.append(Whitespace())
            # Bin
            elif type(value) in value_converter.bin_types:
                line = "{} = (bin)".format(key)
                base.append(context_indent_str + line)
                result = value_converter.bin_encoder(value)
                for item in result:
                    line = context_indent_str + indent_str + item
                    base.append(line)
                #if result:
                line = context_indent_str + indent_str + "---"
                base.append(line)
            # Raw
            elif type(value) in value_converter.raw_types:
                line = "{} = (raw)".format(key)
                base.append(context_indent_str + line)
                result = value_converter.raw_encoder(value)
                for item in result:
                    line = context_indent_str + indent_str + item
                    base.append(line)
                #if result:
                line = context_indent_str + indent_str + "---"
                base.append(line)
            # Text
            elif type(value) in value_converter.text_types:
                line = "{} = (text)".format(key)
                base.append(context_indent_str + line)
                result = value_converter.text_encoder(value)
                for item in result:
                    line = context_indent_str + indent_str + item
                    base.append(line)
                #if result:
                line = context_indent_str + indent_str + "---"
                base.append(line)
            # list box
            elif type(value) in value_converter.list_types:
                misc.update_cached_refs(value, cached_refs)
                result = value_converter.encode(value)
                line = "{} = {}".format(key, result)
                base.append(context_indent_str + line)
                new_context = Context("list", value, context.indents + 1)
                stack.append(new_context)
                DictToBase._loop(value, value_converter, base=base,
                                 stack=stack)
            # dict box
            elif type(value) in value_converter.dict_types:
                misc.update_cached_refs(value, cached_refs)
                result = value_converter.encode(value)
                line = "{} = {}".format(key, result)
                base.append(context_indent_str + line)
                new_context = Context("dict", value, context.indents + 1)
                stack.append(new_context)
                DictToBase._loop(value, value_converter, base=base,
                                 stack=stack)
            # regular value (scalar, datetime, etc)
            else:
                key = key.replace("\n", "\\n")
                result = value_converter.encode(value)
                line = "{} = {}".format(key, result)
                base.append(context_indent_str + line)

    @staticmethod
    def _handle_list(data, context, value_converter, base, stack, cached_refs):
        indent_str = " " * const.INDENT_WIDTH
        context_indent_str = indent_str * context.indents
        for value in data:
            # comments
            if isinstance(value, Comment):
                value = value.replace("\n", "\\n")
                base.append(context_indent_str + value)
            # whitespace
            elif isinstance(value, Whitespace):
                #base.append(context_indent_str + value)
                base.append(Whitespace())
            # Bin
            elif type(value) in value_converter.bin_types:
                line = "(bin)"
                base.append(context_indent_str + line)
                result = value_converter.bin_encoder(value)
                for item in result:
                    line = context_indent_str + context_indent_str + item
                    base.append(line)
                #if result:
                line = context_indent_str + context_indent_str + "---"
                base.append(line)
            # Raw
            elif type(value) in value_converter.raw_types:
                line = "(raw)"
                base.append(context_indent_str + line)
                result = value_converter.raw_encoder(value)
                for item in result:
                    line = context_indent_str + context_indent_str + item
                    base.append(line)
                #if result:
                line = context_indent_str + context_indent_str + "---"
                base.append(line)
            # Text
            elif type(value) in value_converter.text_types:
                line = "(text)"
                base.append(context_indent_str + line)
                result = value_converter.text_encoder(value)
                for item in result:
                    line = context_indent_str + context_indent_str + item
                    base.append(line)
                #if result:
                line = context_indent_str + context_indent_str + "---"
                base.append(line)
            # list box
            elif type(value) in value_converter.list_types:
                misc.update_cached_refs(value, cached_refs)
                result = value_converter.encode(value)
                base.append(context_indent_str + result)
                new_context = Context("list", value, context.indents + 1)
                stack.append(new_context)
                DictToBase._loop(value, value_converter, base=base,
                                 stack=stack)
            # dict box
            elif type(value) in value_converter.dict_types:
                misc.update_cached_refs(value, cached_refs)
                result = value_converter.encode(value)
                base.append(context_indent_str + result)
                new_context = Context("dict", value, context.indents + 1)
                stack.append(new_context)
                DictToBase._loop(value, value_converter, base=base,
                                 stack=stack)
            # regular value (scalar, datetime, etc)
            else:
                result = value_converter.encode(value)
                base.append(context_indent_str + result)


Context = namedtuple("Context", ["name", "collection", "indents"])


class ValueConverter:
    """This class allows the developer to customize value encoding"""
    def __init__(self, dict_type=None, list_type=None,

                 dict_types=None, list_types=None,

                 bin_types=None, bool_types=None,
                 complex_types=None, date_types=None,
                 datetime_types=None, float_types=None,
                 integer_types=None, raw_types=None,
                 string_types=None, text_types=None,
                 time_types=None,

                 bin_encoder=None, bool_encoder=None,
                 complex_encoder=None, date_encoder=None,
                 datetime_encoder=None, float_encoder=None,
                 integer_encoder=None, null_encoder=None,
                 raw_encoder=None, string_encoder=None,
                 text_encoder=None, time_encoder=None,
                 fallback_encoder=None,

                 bin_decoder=None, bool_decoder=None,
                 complex_decoder=None, date_decoder=None,
                 datetime_decoder=None, float_decoder=None,
                 integer_decoder=None, null_decoder=None,
                 raw_decoder=None, string_decoder=None,
                 text_decoder=None, time_decoder=None,
                 fallback_decoder=None):
        """
        All parameters are mirrored with read and write properties

        [parameters]

        - dict_type: the Python type in which a Jesth Dict should be converted into. Defaults to Python dict type.

        - list_type: the Python type in which a Jesth List should be converted into. Defaults to Python list type.

        - XXX_types: this represents a group of parameters. Here, XXX is a placeholder for a Jesth data type. Valid types are: dict, list, bin, bool, complex, date, datetime, float, integer, raw, string, text, time. Examples: dict_types, float_types, and time_types. Use this parameter to set a list of Python types that may be encoded in the Jesth type (the same used as prefix). Example: dict_types defaults to [OrderedDict, dict], i.e., while encoding some Python data, an OrderedDict instance or a regular dict instance will be encoded as a Jesth dict.

        - XXX_encoder: this represents a group of parameters. Here, XXX is a placeholder for one of: bin, bool, complex, date, datetime, float, integer, null, raw, string, text, time, fallback. Use these parameters to set a callable to encode Python values into Jesth values. Example: float_encoder = decimal.Decimal

        - XXX_decoder: this represents a group of parameters. Here, XXX is a placeholder for one of: bin, bool, complex, date, datetime, float, integer, null, raw, string, text, time, fallback. Use these parameters to set a callable to decode Jesth values into Python values. Example: integer_decoder = int
        """

        # default types for dict and list containers
        self._dict_type = dict_type if dict_type else OrderedDict
        self._list_type = list_type if list_type else list

        # types for containers
        self._dict_types = dict_types if dict_types else [OrderedDict, dict]
        self._list_types = list_types if list_types else [list, tuple, set]

        # data types
        self._bin_types = bin_types if bin_types else [bytes]
        self._bool_types = bool_types if bool_types else [bool]
        self._complex_types = complex_types if complex_types else [complex]
        self._date_types = date_types if date_types else [datetime.date]
        self._datetime_types = datetime_types if datetime_types else [datetime.datetime]
        self._float_types = float_types if float_types else [float, Decimal]
        self._integer_types = integer_types if integer_types else [int, HexInt, OctInt, BinInt]
        self._raw_types = raw_types if raw_types else [RawText]
        self._string_types = string_types if string_types else [str, RawString]
        self._text_types = text_types if text_types else [Text]
        self._time_types = time_types if time_types else [datetime.time]

        # encoders
        self._bin_encoder = bin_encoder if bin_encoder else encode_bin
        self._bool_encoder = bool_encoder if bool_encoder else encode_bool
        self._complex_encoder = complex_encoder if complex_encoder else encode_complex
        self._date_encoder = date_encoder if date_encoder else encode_date
        self._datetime_encoder = datetime_encoder if datetime_encoder else encode_datetime
        self._float_encoder = float_encoder if float_encoder else encode_float
        self._integer_encoder = integer_encoder if integer_encoder else encode_integer
        self._null_encoder = null_encoder if null_encoder else encode_null
        self._raw_encoder = raw_encoder if raw_encoder else encode_raw
        self._string_encoder = string_encoder if string_encoder else encode_string
        self._text_encoder = text_encoder if text_encoder else encode_text
        self._time_encoder = time_encoder if time_encoder else encode_time
        self._fallback_encoder = fallback_encoder if fallback_encoder else encode_value

        # decoders
        self._bin_decoder = bin_decoder if bin_decoder else decode_bin
        self._bool_decoder = bool_decoder if bool_decoder else decode_bool
        self._complex_decoder = complex_decoder if complex_decoder else decode_complex
        self._date_decoder = date_decoder if date_decoder else decode_date
        self._datetime_decoder = datetime_decoder if datetime_decoder else decode_datetime
        self._float_decoder = float_decoder if float_decoder else decode_float
        self._integer_decoder = integer_decoder if integer_decoder else decode_integer
        self._null_decoder = null_decoder if null_decoder else decode_null
        self._raw_decoder = raw_decoder if raw_decoder else decode_raw
        self._string_decoder = string_decoder if string_decoder else decode_string
        self._text_decoder = text_decoder if text_decoder else decode_text
        self._time_decoder = time_decoder if time_decoder else decode_time
        self._fallback_decoder = fallback_decoder if fallback_decoder else decode_value

    # default types for containers

    @property
    def bin_type(self):
        return self._bin_type

    @bin_type.setter
    def bin_type(self, val):
        self._bin_type = val

    @property
    def dict_type(self):
        return self._dict_type

    @dict_type.setter
    def dict_type(self, val):
        self._dict_type = val

    @property
    def list_type(self):
        return self._list_type

    @list_type.setter
    def list_type(self, val):
        self._list_type = val

    @property
    def raw_type(self):
        return self._raw_type

    @raw_type.setter
    def raw_type(self, val):
        self._raw_type = val

    @property
    def text_type(self):
        return self._text_type

    @text_type.setter
    def text_type(self, val):
        self._text_type = val

    # types for containers

    @property
    def dict_types(self):
        return self._dict_types

    @dict_types.setter
    def dict_types(self, val):
        self._dict_types = val

    @property
    def list_types(self):
        return self._list_types

    @list_types.setter
    def list_types(self, val):
        self._list_types = val

    # types for values

    @property
    def bin_types(self):
        return self._bin_types

    @bin_types.setter
    def bin_types(self, val):
        self._bin_types = val

    @property
    def bool_types(self):
        return self._bool_types

    @bool_types.setter
    def bool_types(self, val):
        self._bool_types = val

    @property
    def complex_types(self):
        return self._complex_types

    @complex_types.setter
    def complex_types(self, val):
        self._complex_types = val

    @property
    def date_types(self):
        return self._date_types

    @date_types.setter
    def date_types(self, val):
        self._date_types = val

    @property
    def datetime_types(self):
        return self._datetime_types

    @datetime_types.setter
    def datetime_types(self, val):
        self._datetime_types = val

    @property
    def float_types(self):
        return self._float_types

    @float_types.setter
    def float_types(self, val):
        self._float_types = val

    @property
    def integer_types(self):
        return self._integer_types

    @integer_types.setter
    def integer_types(self, val):
        self._integer_types = val

    @property
    def raw_types(self):
        return self._raw_types

    @raw_types.setter
    def raw_types(self, val):
        self._raw_types = val

    @property
    def string_types(self):
        return self._string_types

    @string_types.setter
    def string_types(self, val):
        self._string_types = val

    @property
    def text_types(self):
        return self._text_types

    @text_types.setter
    def text_types(self, val):
        self._text_types = val

    @property
    def time_types(self):
        return self._time_types

    @time_types.setter
    def time_types(self, val):
        self._time_types = val

    # encoders

    @property
    def bin_encoder(self):
        return self._bin_encoder

    @bin_encoder.setter
    def bin_encoder(self, val):
        self._bin_encoder = val

    @property
    def bool_encoder(self):
        return self._bool_encoder

    @bool_encoder.setter
    def bool_encoder(self, val):
        self._bool_encoder = val

    @property
    def complex_encoder(self):
        return self._complex_encoder

    @complex_encoder.setter
    def complex_encoder(self, val):
        self._complex_encoder = val

    @property
    def date_encoder(self):
        return self._date_encoder

    @date_encoder.setter
    def date_encoder(self, val):
        self._date_encoder = val

    @property
    def datetime_encoder(self):
        return self._datetime_encoder

    @datetime_encoder.setter
    def datetime_encoder(self, val):
        self._datetime_encoder = val

    @property
    def dict_encoder(self):
        return self._dict_encoder

    @dict_encoder.setter
    def dict_encoder(self, val):
        self._dict_encoder = val

    @property
    def float_encoder(self):
        return self._float_encoder

    @float_encoder.setter
    def float_encoder(self, val):
        self._float_encoder = val

    @property
    def integer_encoder(self):
        return self._integer_encoder

    @integer_encoder.setter
    def integer_encoder(self, val):
        self._integer_encoder = val

    @property
    def list_encoder(self):
        return self._list_encoder

    @list_encoder.setter
    def list_encoder(self, val):
        self._list_encoder = val

    @property
    def null_encoder(self):
        return self._null_encoder

    @null_encoder.setter
    def null_encoder(self, val):
        self._null_encoder = val

    @property
    def raw_encoder(self):
        return self._raw_encoder

    @raw_encoder.setter
    def raw_encoder(self, val):
        self._raw_encoder = val

    @property
    def string_encoder(self):
        return self._string_encoder

    @string_encoder.setter
    def string_encoder(self, val):
        self._string_encoder = val

    @property
    def text_encoder(self):
        return self._text_encoder

    @text_encoder.setter
    def text_encoder(self, val):
        self._text_encoder = val

    @property
    def time_encoder(self):
        return self._time_encoder

    @time_encoder.setter
    def time_encoder(self, val):
        self._time_encoder = val     
            
    @property
    def fallback_encoder(self):
        return self._fallback_encoder

    @fallback_encoder.setter
    def fallback_encoder(self, val):
        self._fallback_encoder = val

    # decoders

    @property
    def bin_decoder(self):
        return self._bin_decoder

    @bin_decoder.setter
    def bin_decoder(self, val):
        self._bin_decoder = val

    @property
    def bool_decoder(self):
        return self._bool_decoder

    @bool_decoder.setter
    def bool_decoder(self, val):
        self._bool_decoder = val

    @property
    def complex_decoder(self):
        return self._complex_decoder

    @complex_decoder.setter
    def complex_decoder(self, val):
        self._complex_decoder = val

    @property
    def date_decoder(self):
        return self._date_decoder

    @date_decoder.setter
    def date_decoder(self, val):
        self._date_decoder = val

    @property
    def datetime_decoder(self):
        return self._datetime_decoder

    @datetime_decoder.setter
    def datetime_decoder(self, val):
        self._datetime_decoder = val

    @property
    def dict_decoder(self):
        return self._dict_decoder

    @dict_decoder.setter
    def dict_decoder(self, val):
        self._dict_decoder = val

    @property
    def float_decoder(self):
        return self._float_decoder

    @float_decoder.setter
    def float_decoder(self, val):
        self._float_decoder = val

    @property
    def integer_decoder(self):
        return self._integer_decoder

    @integer_decoder.setter
    def integer_decoder(self, val):
        self._integer_decoder = val

    @property
    def list_decoder(self):
        return self._list_decoder

    @list_decoder.setter
    def list_decoder(self, val):
        self._list_decoder = val

    @property
    def null_decoder(self):
        return self._null_decoder

    @null_decoder.setter
    def null_decoder(self, val):
        self._null_decoder = val

    @property
    def raw_decoder(self):
        return self._raw_decoder

    @raw_decoder.setter
    def raw_decoder(self, val):
        self._raw_decoder = val

    @property
    def string_decoder(self):
        return self._string_decoder

    @string_decoder.setter
    def string_decoder(self, val):
        self._string_decoder = val

    @property
    def text_decoder(self):
        return self._text_decoder

    @text_decoder.setter
    def text_decoder(self, val):
        self._text_decoder = val

    @property
    def time_decoder(self):
        return self._time_decoder

    @time_decoder.setter
    def time_decoder(self, val):
        self._time_decoder = val

    @property
    def fallback_decoder(self):
        return self._fallback_decoder

    @fallback_decoder.setter
    def fallback_decoder(self, val):
        self._fallback_decoder = val

    def encode(self, value):
        # box
        container_tag = self._encode_container(value)
        if container_tag:
            return container_tag
        # null
        if value is None:
            return self._null_encoder(value)
        # == data types
        value_type = type(value)
        # bool
        if value_type in self._bool_types:
            return self._bool_encoder(value)
        # complex
        if value_type in self._complex_types:
            return self._complex_encoder(value)
        # date
        if value_type in self._date_types:
            return self._date_encoder(value)
        # datetime
        if value_type in self._datetime_types:
            return self._datetime_encoder(value)
        # float
        if value_type in self._float_types:
            return self._float_encoder(value)
        # integer
        if value_type in self._integer_types:
            return self._integer_encoder(value)
        # string
        if value_type in self._string_types:
            return self._string_encoder(value)
        # time
        if value_type in self._time_types:
            return self._time_encoder(value)
        # default
        return self._fallback_encoder(value)

    def decode(self, value):
        value = value.strip()
        container = self._decode_container(value)
        if container is not None:
            return container
        # scalars and co.  (note that the order of items matters here)
        decoders = (self._null_decoder, self._bool_decoder,
                    self._string_decoder, self._complex_decoder,
                    self._integer_decoder, self._float_decoder,
                    self._date_decoder, self._time_decoder,
                    self._datetime_decoder, self._fallback_decoder)
        for p in decoders:
            try:
                result = p(value)
            except Exception as e:
                continue
            else:
                return result
        raise error.ConversionError

    def _encode_container(self, val):
        value_type = type(val)
        # bin
        if value_type in self._bin_types:
            return "(bin)"
        # dict
        if value_type in self._dict_types:
            return "(dict)"
        # list
        if value_type in self._list_types:
            return "(list)"
        # raw
        if value_type in self._raw_types:
            return "(raw)"
        # text
        if value_type in self._text_types:
            return "(text)"
        return None

    def _decode_container(self, val):
        if not val.startswith("(") or not val.endswith(")"):
            return None
        # bin
        if val == "(bin)":
            return list()
        # dict
        if val == "(dict)":
            return self._dict_type()
        # list
        if val == "(list)":
            return self._list_type()
        # raw
        if val == "(raw)":
            return list()
        # text
        if val == "(text)":
            return list()
        msg = "Expected one of these containers: (bin), (dict), (list), (raw), (text)"
        raise error.Error(msg)


# === Encoders ===

def encode_bin(val):
    if not val:
        return list()
    r = base64.standard_b64encode(val)
    r = r.decode("utf-8")
    r = r.replace("\n", "")
    result = list()
    cache = list()
    i = 0
    for char in r:
        if i == 76:
            i = 0
            result.append("".join(cache))
            cache = list()
            cache.append(char)
            continue
        cache.append(char)
        i += 1
    if cache:
        result.append("".join(cache))
    return result
    

def encode_bool(val):
    if val:
        return "true"
    return "false"


def encode_complex(val):
    real, imag = val.real, val.imag
    real = int(real) if real.is_integer() else real
    imag = int(imag) if imag.is_integer() else imag
    tidy_real = misc.tidy_up_float(real)
    tidy_imag = misc.tidy_up_float(imag)
    return "{}{}{}i".format(tidy_real, "" if imag < 0 else "+", tidy_imag)


def encode_date(val):
    """Parse date and time with the ISO 8601 format"""
    # date ISO format
    return str(val)


def encode_datetime(val):
    value = val.isoformat()
    if value.endswith("+00:00"):
        value = value.replace("+00:00", "Z")
    return str(value)


def encode_float(val):
    if isinstance(val, Decimal):
        val = "{:E}".format(val)
    return misc.tidy_up_float(val)


def encode_integer(val):
    t = type(val)
    if t is int:
        return misc.tidy_up_int(val)
    if t is HexInt:
        return misc.tidy_up_hex(hex(val))
    if t is OctInt:
        return misc.tidy_up_oct(oct(val))
    if t is BinInt:
        return misc.tidy_up_bin(bin(val))
    return str(val)


def encode_null(val):
    return "null"


def encode_raw(val):
    if not val:
        return list()
    return val.split("\n")


def encode_string(val):
    t = type(val)
    # raw string (literal)
    if t is RawString:
        val = val.replace("\n", "\\n")
        return "'{}'".format(val)
    # processed string
    val = val.replace("\\", "\\\\")
    val = val.replace("\n", "\\n")
    return '"{}"'.format(val)


def encode_text(val):
    if not val:
        return list()
    val = val.replace("\\", "\\\\")
    return val.split("\n")


def encode_time(val):
    # time ISO format
    return str(val)


def encode_value(val):
    return encode_string(str(val))


# === Decoders ===


def decode_bin(val):
    # list of strings
    val = "".join(val)
    return base64.standard_b64decode(val)


def decode_bool(val):
    val = val.lower()
    if val == "true":
        return bool(1)
    elif val == "false":
        return bool(0)
    raise error.ConversionError


def decode_complex(val):
    if val.endswith("i"):
        val = val.replace(" ", "")
        val = val.rstrip("i") + "j"
        return complex(val)
    raise error.ConversionError


def decode_date(val):
    """Parse date and time with the ISO 8601 format"""
    # date ISO format
    return datetime.date.fromisoformat(val)


def decode_datetime(val):
    # datetime (2023-03-13T09:10:42Z)
    if val.endswith("Z"):
        val += "+00:00"
    return datetime.datetime.fromisoformat(val)


def decode_float(val):
    if "e" in val.lower():
        return Decimal(val)
    if "." in val:
        return float(val)
    raise error.ConversionError


def decode_integer(val):
    v = val.lstrip("+-")
    # hexadecimal
    if v.startswith("0x"):
        val = int(val, 16)
        return HexInt(val)
    # octal
    if v.startswith("0o"):
        val = int(val, 8)
        return OctInt(val)
    # binary
    if v.startswith("0b"):
        val = int(val, 2)
        return BinInt(val)
    # default base 10
    return int(val)


def decode_null(val):
    val = val.lower()
    if val == "null":
        return None
    raise error.ConversionError


def decode_raw(val):
    # val is a list of string
    text = "\n".join(val)
    return RawText(text)


def decode_string(val):
    if val.startswith('"') and val.endswith('"'):
        return misc.decode_unicode(val[1:-1])
    if val.startswith("'") and val.endswith("'"):
        return RawString(val[1:-1])
    raise error.ConversionError
    #val = misc.replace_newlines(val)
    #val = misc.replace_tabs(val)


def decode_text(val):
    text = "\n".join(val)
    text = misc.decode_unicode(text)
    return Text(text)


def decode_time(val):
    # time ISO format
    return datetime.time.fromisoformat(val)


def decode_value(val):
    raise error.ConversionError



import unittest
import os.path
import tempfile
from jesth import misc, error, const
from jesth.section import Section


class TestEnsureParentDirFunc(unittest.TestCase):

    def setUp(self):
        self._temp_dir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self._temp_dir.cleanup()

    def test(self):
        path1 = os.path.join(self._temp_dir.name, "sub", "dir")
        path2 = os.path.join(path1, "file.txt")
        self.assertEqual(False, os.path.isdir(path1))
        misc.ensure_parent_dir(path2)
        self.assertEqual(True, os.path.isdir(path1))


class TestSplitValueFunc(unittest.TestCase):

    def test_with_default_args(self):
        line = " key = 'value here' "
        r = misc.split_key_value(line)
        expected = ("key", "'value here'")
        self.assertEqual(expected, r)

    def test_with_whitespace_stripping_disabled(self):
        line = " key = 'value here' "
        r = misc.split_key_value(line, strip_whitespace=False)
        expected = (" key ", " 'value here' ")
        self.assertEqual(expected, r)

    def test_with_colon_as_separator(self):
        line = " key : 'value here' "
        r = misc.split_key_value(line, sep=":")
        expected = ("key", "'value here'")
        self.assertEqual(expected, r)


class TestUpdateCachedRefsFunc(unittest.TestCase):

    def test(self):
        cached_refs = list()
        misc.update_cached_refs(1, cached_refs)
        self.assertEqual(1, len(cached_refs))
        with self.assertRaises(error.Error):
            misc.update_cached_refs(1, cached_refs)


class TestCountIndentsFunc(unittest.TestCase):

    def test_with_valid_indent(self):
        indent = " " * const.INDENT_WIDTH
        hello_world = "Hello World !"
        line = indent + indent + hello_world
        r = misc.count_indents(line)
        expected = 2
        self.assertEqual(expected, r)

    def test_with_invalid_indent(self):
        indent = " " * (const.INDENT_WIDTH + 1)
        hello_world = "Hello World !"
        line = indent + indent + hello_world
        with self.assertRaises(error.IndentError):
            misc.count_indents(line)


class TestCleanLeadingBackslashesFunc(unittest.TestCase):

    def test_trivial_line(self):
        leading_backslash = "\\"
        line1 = "Hello World !"
        line2 = leading_backslash + line1
        r = misc.clean_leading_backslashes(line2)
        expected = leading_backslash + line1
        self.assertEqual(expected, r)

    def test_line_without_leading_backslash(self):
        line = "[Hello World !"
        r = misc.clean_leading_backslashes(line)
        expected = line
        self.assertEqual(expected, r)

    def test_line_with_one_leading_backslash(self):
        leading_backslash = "\\"
        line1 = "[Hello World !"
        line2 = leading_backslash + line1
        r = misc.clean_leading_backslashes(line2)
        expected = line1
        self.assertEqual(expected, r)

    def test_line_with_two_leading_backslash(self):
        leading_backslash = "\\"
        line1 = "[Hello World !"
        line2 = leading_backslash + leading_backslash + line1
        r = misc.clean_leading_backslashes(line2)
        expected = leading_backslash + line1
        self.assertEqual(expected, r)

    def test_line_with_three_leading_backslash(self):
        leading_backslash = "\\"
        line1 = "[Hello World !"
        line2 = leading_backslash + leading_backslash + leading_backslash + line1
        r = misc.clean_leading_backslashes(line2)
        expected = leading_backslash + leading_backslash + line1
        self.assertEqual(expected, r)


class TestAddLeadingBackslashesFunc(unittest.TestCase):

    def test_trivial_line(self):
        line = "Hello World !"
        r = misc.add_leading_backslashes(line)
        expected = line
        self.assertEqual(expected, r)

    def test_candidate_line(self):
        leading_backslash = "\\"
        line = "[Hello World !"
        r = misc.add_leading_backslashes(line)
        expected = leading_backslash + line
        self.assertEqual(expected, r)

    def test_candidate_line_with_leading_backslash(self):
        leading_backslash = "\\"
        line = leading_backslash + "[Hello World !"
        r = misc.add_leading_backslashes(line)
        expected = leading_backslash + line
        self.assertEqual(expected, r)


class TestGetHeadersFunc(unittest.TestCase):

    def test(self):
        section = Section("")
        section1 = Section("header1")
        section2 = Section("header2")
        sections = (section, section1, section2)
        r = misc.get_headers(sections)
        expected = {"", "header1", "header2"}
        self.assertEqual(expected, r)


class TestDecodeUnicodeFunc(unittest.TestCase):

    def test_string_1(self):
        s = r"hello \n \\n world \u02eb"
        r = misc.decode_unicode(s)
        expected = "hello \n \\n world \u02eb"
        self.assertEqual(expected, r)

    def test_string_2(self):
        s = r"hello \n world \u02eb ˫"
        r = misc.decode_unicode(s)
        expected = "hello \n world \u02eb ˫"
        self.assertEqual(expected, r)

    def test_string_3(self):
        s = r"hélicoptère"
        r = misc.decode_unicode(s)
        expected = "hélicoptère"
        self.assertEqual(expected, r)

    def test_string_4(self):
        s = r"flag for Japan \U0001f1ef\U0001f1f5"
        r = misc.decode_unicode(s)
        expected = "flag for Japan 🇯🇵"
        self.assertEqual(expected, r)

    def test_string_5(self):
        s = r"flag for Japan \U0001f1ef\U0001f1f5"
        r = misc.decode_unicode(s)
        expected = "flag for Japan \U0001f1ef\U0001f1f5"
        self.assertEqual(expected, r)


class TestEncodeUnicodeFunc(unittest.TestCase):

    def test_string_1(self):
        s = "hello \n \\n world \u02eb"
        r = misc.encode_unicode(s)
        expected = "hello \n \\n world \\u02eb"
        self.assertEqual(expected, r)

    def test_string_2(self):
        s = "hello \n\u000a world \u02eb ˫"
        r = misc.encode_unicode(s)
        expected = "hello \n\n world \\u02eb \\u02eb"
        self.assertEqual(expected, r)

    def test_string_3(self):
        s = "hélicoptère"
        r = misc.encode_unicode(s)
        expected = "hélicoptère"
        self.assertEqual(expected, r)

    def test_string_4(self):
        s = "flag for Japan \U0001f1ef\U0001f1f5"
        r = misc.encode_unicode(s)
        expected = "flag for Japan \\U0001f1ef\\U0001f1f5"
        self.assertEqual(expected, r)

    def test_string_5(self):
        s = "flag for Japan 🇯🇵"
        r = misc.encode_unicode(s)
        expected = "flag for Japan \\U0001f1ef\\U0001f1f5"
        self.assertEqual(expected, r)


class TestTidyUpNumberFuncs(unittest.TestCase):

    def test_tidy_up_integer(self):
        x = -5000000
        r = misc.tidy_up_int(x)
        expected = "-5_000_000"
        self.assertEqual(expected, r)

    def test_tidy_up_float(self):
        x = "-3780678.146538980757E-5"
        r = misc.tidy_up_float(x)
        expected = "-3_780_678.146_538_980_757E-5"
        self.assertEqual(expected, r)

    def test_tidy_up_hex(self):
        x = "-0x5A0546E8F"
        r = misc.tidy_up_hex(x)
        expected = "-0x5_A054_6E8F"
        self.assertEqual(expected, r)

    def test_tidy_up_bin(self):
        x = "-0b10110011101"
        r = misc.tidy_up_bin(x)
        expected = "-0b101_1001_1101"
        self.assertEqual(expected, r)

    def test_tidy_up_oct(self):
        x = "-0o2345623366"
        r = misc.tidy_up_oct(x)
        expected = "-0o23_4562_3366"
        self.assertEqual(expected, r)


class TestParseFloatFunc(unittest.TestCase):

    def test(self):
        x = "-3.14653898075720390738E-5"
        info = misc.parse_float(x)
        expected_left_mantissa = "-3"
        expected_right_mantissa = "14653898075720390738"
        expected_exponent = "-5"
        self.assertEqual(expected_left_mantissa, info.left_mantissa)
        self.assertEqual(expected_right_mantissa, info.right_mantissa)
        self.assertEqual(expected_exponent, info.exponent)


if __name__ == "__main__":
    unittest.main()



import unittest
import tempfile
import pathlib
from jesth.parser import parse
from jesth.renderer import render, write


TEXT = """\
[section1]
first
section


[section2]
second
section\
"""


class TestRenderFunc(unittest.TestCase):

    def test_render_all_sections(self):
        document = parse(TEXT)
        r = render(document, spacing=2)
        expected = TEXT
        self.assertEqual(expected, r)

    def test_render_only_one_section(self):
        document = parse(TEXT)
        r = render(document, "section2")
        expected = "[section2]\nsecond\nsection"
        self.assertEqual(expected, r)


class TestWriteFunc(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.close()
        self._path_str = file.name
        self._path_obj = pathlib.Path(self._path_str)
        self._document = parse(TEXT)

    def tearDown(self):
        self._path_obj.unlink()

    def test_write_to_path_obj(self):
        expected = write(self._document, self._path_obj)
        _check_file_contents(self, self._path_str, expected)

    def test_write_to_path_str(self):
        expected = write(self._document, self._path_str)
        _check_file_contents(self, self._path_str, expected)

    def test_write_to_file(self):
        with open(self._path_str, "wb") as file:
            expected = write(self._document, file)
        _check_file_contents(self, self._path_str, expected)


def _check_file_contents(test_case, path_str, expected_contents):
    with open(path_str, "r") as file:
        contents = file.read()
        test_case.assertEqual(expected_contents, contents)


if __name__ == '__main__':
    unittest.main()



import decimal
import unittest
import datetime
from decimal import Decimal
from jesth import box, error
from jesth.converter import ValueConverter, make_dict, flatten_dict


BODY = r"""
# strings
string = "Hello "World" ! \\ \n \\n \\u000A ˫"
raw_string = 'Hello 'World' ! \n \\n \u000A ˫'
empty_string = ""
empty_raw_string = ''

# scalars
int = -4_200_000
hex_int = 0xAE1F
oct_int = 0o3427
bin_int = 0b101_0011_1010_1010
float = 3.141
decimal_float = 3.141_592_653_589_793_238_46E-10
complex = -4.2+6.9i
bool_1 = true
bool_2 = false

# null value
empty_value = null

# date and time (ISO 8601)
datetime = 2020-10-20T15:35:57Z
date = 2020-10-20
time = 15:35:57

# text
text = (text)
    Stand a little less \\n
    between me and the sun. ˫
    ---
empty_text = (text)
    ---

# raw text
raw = (raw)
    The foundation of every state
    is the education of its youth. ˫
    \u000A - WYSIWYG - C:\home\alex
    ---
empty_raw = (raw)
    ---

# bin data (standard base64 - RFC 4648)
bin = (bin)
    TG9yZW0gaXBzdW0gZG9sb3Igc2l0IGFtZXQsIGNvbnNlY3RldHVyIGFkaXBpc2NpbmcgZWxpdCwg
    c2VkIGRvIGVpdXNtb2QgdGVtcG9yIGluY2lkaWR1bnQgdXQgbGFib3JlIGV0IGRvbG9yZSBtYWduY
    SBhbGlxdWEu
    ---
empty_bin = (bin)
    ---

# list collection
list = (list)
    "Item 1"
    "Item 2"
    "Item 3"
    # nested list
    (list)
        "item i"
        "item ii"
        "item iii"
empty_list = (list)

# dict collection
dict = (dict)
    key_1 = "value"
    key_2 = (list)
        "item i"
        "item ii"
        # nested dict
        (dict)
            project = "Jesth: Just Extract Sections Then Hack !"
            description = "Next level human-readable data serialization format"
            repository = 'https://github.com/pyrustic/jesth'
            author = "alex rustic"
empty_dict = (dict)
"""


class TestConversionFuncs(unittest.TestCase):

    def test(self):
        # Make a dict from BODY
        dict_body = make_dict(BODY, strict=False)
        # Flatten the dict
        flat_body = flatten_dict(dict_body)
        # Make text from flatten body then compare with BODY
        text = "\n".join(flat_body)
        self.assertEqual(BODY, text)


class TestValueConverterClass(unittest.TestCase):

    def setUp(self):
        self._value_converter = ValueConverter()

    def test_bin_encoder(self):
        val = b"Hello World"
        r = self._value_converter.encode(val)
        expected = "(bin)"
        self.assertEqual(expected, r)

    def test_bool_encoder(self):
        # test True
        val = True
        r = self._value_converter.encode(val)
        expected = "true"
        self.assertEqual(expected, r)
        # test False
        val = False
        r = self._value_converter.encode(val)
        expected = "false"
        self.assertEqual(expected, r)

    def test_complex_encoder(self):
        val = 3.14 + 5j
        r = self._value_converter.encode(val)
        expected = "3.14+5i"
        self.assertEqual(expected, r)

    def test_date_encoder(self):
        val = datetime.date(2023, 5, 1)
        r = self._value_converter.encode(val)
        expected = "2023-05-01"
        self.assertEqual(expected, r)

    def test_datetime_encoder(self):
        # Datetime 1 (datetime with T between date and time)
        val = datetime.datetime(2023, 5, 1, 8, 10, 32,
                                tzinfo=datetime.timezone.utc)
        r = self._value_converter.encode(val)
        expected = "2023-05-01T08:10:32Z"
        self.assertEqual(expected, r)
        # Datetime 3 (datetime with +8 hours offset)
        timezone = datetime.timezone(datetime.timedelta(hours=8))
        val = datetime.datetime(2023, 5, 1, 8, 10, 32,
                                tzinfo=timezone)
        r = self._value_converter.encode(val)
        expected = "2023-05-01T08:10:32+08:00"
        self.assertEqual(expected, r)

    def test_float_encoder(self):
        val = 3.14E2
        r = self._value_converter.encode(val)
        expected = "314.0"
        self.assertEqual(expected, r)

    def test_custom_float_encoder(self):
        val = Decimal("3.148739534274212841284649")
        r = self._value_converter.encode(val)
        expected = "3.148_739_534_274_212_841_284_649E0"
        self.assertEqual(expected, r)

    def test_integer_encoder(self):
        # base 10 integer
        val = -123456789
        r = self._value_converter.encode(val)
        expected = "-123_456_789"
        self.assertEqual(expected, r)
        # hex integer
        val = box.HexInt(0x13FA89D23E9)
        r = self._value_converter.encode(val)
        expected = "0x13F_A89D_23E9"
        self.assertEqual(expected, r)
        # bin integer
        val = box.BinInt(0b1001011011100)
        r = self._value_converter.encode(val)
        expected = "0b1_0010_1101_1100"
        self.assertEqual(expected, r)
        # oct integer
        val = box.OctInt(0o3727352737251)
        r = self._value_converter.encode(val)
        expected = "0o3_7273_5273_7251"
        self.assertEqual(expected, r)

    def test_null_encoder(self):
        val = None
        r = self._value_converter.encode(val)
        expected = "null"
        self.assertEqual(expected, r)

    def test_raw_encoder(self):
        val = box.RawText()
        r = self._value_converter.encode(val)
        expected = "(raw)"
        self.assertEqual(expected, r)

    def test_string_encoder(self):
        # Processed string
        val = "Hello \n \\n \n World ! \u02eb"
        r = self._value_converter.encode(val)
        expected = r'"Hello \n \\n \n World ! ˫"'
        self.assertEqual(expected, r)
        # Raw string
        val = box.RawString("Raw \n \\n \u000A \\\\n String ! ˫")
        r = self._value_converter.encode(val)
        expected = r"'Raw \n \n \n \\n String ! ˫'"
        self.assertEqual(expected, r)

    def test_text_encoder(self):
        val = box.Text()
        r = self._value_converter.encode(val)
        expected = "(text)"
        self.assertEqual(expected, r)

    def test_time_encoder(self):
        val = datetime.time(10, 30, 58)
        r = self._value_converter.encode(val)
        expected = "10:30:58"
        self.assertEqual(expected, r)

    def test_fallback_encoder(self):
        val = CustomClass()
        r = self._value_converter.encode(val)
        expected = '"My Custom Class"'
        self.assertEqual(expected, r)

    def test_bin_decoder(self):
        val = "(bin)"
        r = self._value_converter.decode(val)
        expected = list()
        self.assertEqual(expected, r)

    def test_bool_decoder(self):
        # test True
        val = "true"
        r = self._value_converter.decode(val)
        expected = True
        self.assertEqual(expected, r)
        # test False
        val = "false"
        r = self._value_converter.decode(val)
        expected = False
        self.assertEqual(expected, r)

    def test_complex_decoder(self):
        val = "3.14+5i"
        r = self._value_converter.decode(val)
        expected = 3.14+5j
        self.assertEqual(expected, r)

    def test_date_decoder(self):
        val = "2023-05-01"
        r = self._value_converter.decode(val)
        expected = datetime.date(2023, 5, 1)
        self.assertEqual(expected, r)

    def test_datetime_decoder(self):
        # Datetime 1 (datetime with T between date and time)
        val = "2023-05-01T08:10:32Z"
        r = self._value_converter.decode(val)
        expected = datetime.datetime(2023, 5, 1, 8, 10, 32,
                                     tzinfo=datetime.timezone.utc)
        self.assertEqual(expected, r)
        # Datetime 2 (datetime with space between date and time)
        val = "2023-05-01 08:10:32Z"
        r = self._value_converter.decode(val)
        self.assertEqual(expected, r)
        # Datetime 3 (datetime with +8 hours offset)
        val = "2023-05-01 08:10:32+08:00"
        r = self._value_converter.decode(val)
        timezone = datetime.timezone(datetime.timedelta(hours=8))
        expected = datetime.datetime(2023, 5, 1, 8, 10, 32,
                                     tzinfo=timezone)
        self.assertEqual(expected, r)

    def test_float_decoder(self):
        # float 1
        val = "3.14E9"
        r = self._value_converter.decode(val)
        expected_type = decimal.Decimal
        expected_r = decimal.Decimal("3.14E9")
        self.assertIs(expected_type, type(r))
        self.assertEqual(expected_r, r)
        # float 2
        val = "3.1437"
        r = self._value_converter.decode(val)
        expected_type = float
        expected_r = 3.1437
        self.assertIs(expected_type, type(r))
        self.assertEqual(expected_r, r)
        # float 3
        val = "3.1437E0"
        r = self._value_converter.decode(val)
        expected_type = decimal.Decimal
        expected_r = decimal.Decimal("3.1437E0")
        self.assertIs(expected_type, type(r))
        self.assertEqual(expected_r, r)

    def test_custom_float_decoder(self):
        self._value_converter.float_decoder = Decimal
        val = "3.148739534274212841284649"
        r = self._value_converter.decode(val)
        expected = Decimal("3.148739534274212841284649")
        self.assertEqual(expected, r)

    def test_integer_decoder(self):
        # base 10 integer
        val = "-123_456_789"
        r = self._value_converter.decode(val)
        expected = -123456789
        self.assertEqual(expected, r)
        self.assertTrue(isinstance(r, int))
        # hex integer
        val = "0x13F_A89D_23E9"
        r = self._value_converter.decode(val)
        expected = 0x13F_A89D_23E9
        self.assertEqual(expected, r)
        self.assertTrue(isinstance(r, box.HexInt))
        # bin integer
        val = "0b1_0010_1101_1100"
        r = self._value_converter.decode(val)
        expected = 0b1_0010_1101_1100
        self.assertEqual(expected, r)
        self.assertTrue(isinstance(r, box.BinInt))
        # oct integer
        val = "0o3_7273_5273_7251"
        r = self._value_converter.decode(val)
        expected = 0o3_7273_5273_7251
        self.assertEqual(expected, r)
        self.assertTrue(isinstance(r, box.OctInt))

    def test_null_decoder(self):
        val = "null"
        r = self._value_converter.decode(val)
        expected = None
        self.assertEqual(expected, r)

    def test_raw_decoder(self):
        val = "(raw)"
        r = self._value_converter.decode(val)
        expected = list()
        self.assertEqual(expected, r)

    def test_string_decoder(self):
        # Processed string
        val = r'"Hello \n \\n \u000A World ! ˫"'
        r = self._value_converter.decode(val)
        expected = "Hello \n \\n \n World ! \u02eb"
        self.assertEqual(expected, r)
        # Raw string
        val = r"'Hello \n \\n \u000A World ! ˫'"
        r = self._value_converter.decode(val)
        expected = "Hello \\n \\\\n \\u000A World ! ˫"
        self.assertEqual(expected, r)

    def test_text_decoder(self):
        val = "(text)"
        r = self._value_converter.decode(val)
        expected = list()
        self.assertEqual(expected, r)

    def test_time_decoder(self):
        val = "10:30:58"
        r = self._value_converter.decode(val)
        expected = datetime.time(10, 30, 58)
        self.assertEqual(expected, r)

    def test_fallback_decoder(self):
        val = "goto"
        with self.assertRaises(error.ConversionError):
            self._value_converter.decode(val)


class CustomClass:
    def __str__(self):
        return "My Custom Class"


if __name__ == '__main__':
    unittest.main()



"""This module exposes read and parse functions that return a document instance"""
import pathlib
from jesth import misc
from jesth.converter import ValueConverter
from jesth.document import Document
from jesth.section import Section


def read(source, value_converter=None):
    """
    Read a file, returns a document (jesth.document.Document)

    [parameters]
    - source: source is either a path string or a pathlib.Path object,
     or a binary file-like object
    - value_converter: instance of jesth.converter.ValueConverter to customize value encoding

    [return]
    Returns an instance of jesth.document.Document
    """
    value_converter = value_converter if value_converter else ValueConverter()
    if isinstance(source, pathlib.Path):
        source = str(source.resolve())
    parser = Parser(value_converter=value_converter)
    if isinstance(source, str):
        with open(source, "rb") as file:
            _read_file_object(file, parser)
    else:
        _read_file_object(source, parser)
    document = Document(source, sections=parser.sections,
                        value_converter=value_converter)
    return document


def parse(text, value_converter=None):
    """
    Parse a text, return a document

    [parameters]
    - text: text string
    - value_converter: instance of jesth.converter.ValueConverter to customize value encoding

    [return]
    Returns an instance of jesth.document.Document
    """
    value_converter = value_converter if value_converter else ValueConverter()
    text = text.split("\n")
    parser = Parser(value_converter=value_converter)
    for line in text:
        if not parser.feed(line):
            break
    document = Document(sections=parser.sections,
                        value_converter=value_converter)
    return document


def _read_file_object(file, parser):
    while True:
        line = file.readline()
        if not line:
            break
        line = line.decode("utf-8").rstrip("\n")
        if not parser.feed(line):
            break


class Parser:
    def __init__(self, value_converter=None):
        self._value_converter = value_converter if value_converter else ValueConverter()
        self._sections = list()
        self._header = None
        self._body = None
        self._feedable = True

    @property
    def value_converter(self):
        return self._value_converter

    @property
    def feedable(self):
        """Isn't anymore feedable when the `[[END]]` tag is encountered"""
        return self._feedable

    @property
    def sections(self):
        self._update()
        return self._sections

    def feed(self, line):
        #if not line or line.isspace():
        #    return
        if not self._feedable:
            return False
        if line.rstrip().lower() == "[[end]]":
            self._update()
            self._feedable = False
            return False
        if line.startswith("["):
            cache = line.rstrip()
            if cache.endswith("]"):
                self._update()
                self._header = cache.strip("[]")
                return True
        line = misc.clean_leading_backslashes(line)
        if self._body is None:
            self._body = list()
        self._body.append(line)
        return True

    def _update(self):
        if not self._feedable:
            return
        if self._header is None and self._body is None:
            return
        if self._header is None and self._body is not None:
            header, body = "", self._body
        elif self._header is not None and self._body is None:
            header, body = self._header, list()
        else:
            header, body = self._header, self._body
        self._strip_section_spacing(body)
        section = Section(header, body, value_converter=self._value_converter)
        self._sections.append(section)
        self._header = self._body = None

    def _strip_section_spacing(self, body):
        for line in reversed(body):
            if line and not line.isspace():
                return
            del body[-1]



"""Errors classes"""


class Error(Exception):
    """Root error class for the Jesth Python library"""
    pass


class ConversionError(Error):
    """Error class for issues during data conversion"""
    pass


class IndentError(Error):
    """Indent error while parsing a section body"""
    pass



"""Document class for creating model for Jesth data or to interacting with a Jesthfile"""
import pathlib
from collections import OrderedDict
from jesth import misc
from jesth.section import Section
from jesth.renderer import write, render
from jesth.converter import ValueConverter


class Document:
    """Create a model for Jesth data or for interacting with a JesthFile"""
    def __init__(self, path=None, sections=None, *, value_converter=None):
        """
        Init the document

        [parameters]
        - path: if the document is linked to a JesthFile,`path` contains the path to this file.
            Path might also be a pathlib.Path instance.
        - sections: list of sections (or tuple)
        - value_converter: an instance of jesth.converter.ValueConverter
        """
        self._path = str(path.resolve()) if isinstance(path, pathlib.Path) else path
        sections = sections if sections else list()
        self._sections = list(sections) if not isinstance(sections, list) else sections
        self._value_converter = value_converter if value_converter else ValueConverter()
        self._is_new = False
        self._model = None
        self._setup()

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, val):
        self._path = val

    @property
    def sections(self):
        return self._sections.copy()

    @sections.setter
    def sections(self, val):
        self._sections = val
        self._create_model()

    @property
    def value_converter(self):
        return self._value_converter

    @property
    def headers(self):
        if not self._sections:
            return None
        return misc.get_headers(self._sections)

    def render(self, *headers, spacing=1):
        """
        Render the entire document or a specific set of sections, i.e.,
        returns a Jesth string that may be stored in a file.

        [parameters]
        - *headers: Headers of sections to render. Omitting this will render the entire section
        - spacing: number of empty lines between two adjacent sections. Defaults to 1 empty line.

        [return]
        Returns a string that contains sections (each made of square-brackets delimited header
        and associated body)
        """
        return render(self, *headers,
                      spacing=spacing)

    def save(self):
        """
        Save the recent modifications if this document is linked to a JesthFile,
         i.e., the path parameter has been set.

        [return]
        Returns True or False
         """
        if not self._path:
            return False
        write(self, destination=self._path)
        return True

    def save_to(self, path):
        """
        Save the document in a specific filename

        [parameters]
        - path: path to filename. Path may be a pathlib.Path instance
        """
        if isinstance(path, pathlib.Path):
            path = str(path.resolve())
        write(self, destination=path)

    def append(self, header, body=None):
        """
        Create a new section then append it to the end of this document

        [parameters]
        - header: the header (string) of the new section
        - body: the body of this section as a string or a list of strings
        """
        self._insert(len(self._sections), header, body)

    def insert(self, index, header, body=None):
        """
        Create a new section then insert it in the document at a specific index

        [parameters]
        - index: integer index (absolute index). Accepts -1 to mimic the "append" method
        - header: the header (string) of the new section
        - body: the body of this section as a string or a list of strings
        """
        index = len(self._sections) if index == -1 else index
        self._insert(index, header, body)

    def get(self, header, index=-1):
        """
        Get X named section object located at Y index relatively to others sections with same header

        [parameters]
        - header: the header (string) of the section
        - index: integer index, relatively to the section
        family (sections sharing this same header).
        Defaults to -1, thus will be returned, the last section with this header relatively
        to this header family.

        [return]
        Returns a section or None
        """
        if header not in self._model:
            return None
        return self._model[header][index]

    def get_all(self, header):
        """
        Get all sections sharing same header

        [parameters]
        - header: the header (string) of the sections

        [return]
        Returns a list of sections whose headers match with the `header` parameter.
        """
        return self._model.get(header, list()).copy()

    def count(self, header):
        """
        Count the number of sections whose headers match with the
        `header` parameter

        [parameters]
        - header: the header (string) of the sections

        [return]
        Returns the integer number of sections matching with the `header` parameter
        """
        if header not in self._model:
            return 0
        return len(self._model[header])

    def remove(self, header, index=-1):
        """
        Remove a section from this document

        [parameters]
        - header: the header of the section to remove
        - index: the index (integer) of the section relatively to
        its family (sections sharing same header).
        Defaults to -1, thus the last section of the given header family will be removed
        from the document
        """
        if header not in self._model:
            return
        if self._model[header]:
            del self._model[header][index]
        else:
            del self._model[header]
        index = len(self._sections) if index == -1 else index
        i = len(self._sections) - 1
        for _ in reversed(self._sections):
            if i == index:
                del self._sections[i]
                break

    def remove_all(self, header):
        """
        Remove all sections with this specific header

        [parameters]
        - header: the header of the section to remove
        """
        if header not in self._model:
            return
        del self._model[header]
        i = len(self._sections) - 1
        for section in reversed(self._sections):
            if section.header == header:
                del self._sections[i]
            i -= 1

    def _setup(self):
        self._create_model()

    def _create_model(self):
        self._model = OrderedDict()
        for section in self._sections:
            header = section.header
            if header not in self._model:
                self._model[header] = list()
            self._model[header].append(section)

    def _insert(self, index, header, body):
        section = Section(header, body=body,
                          value_converter=self._value_converter)
        if index == 0 or index == len(self._sections):
            self._sections.insert(index, section)
            self._update_model(header, index, section)
        else:
            self._sections.insert(index, section)
            self._create_model()

    def _update_model(self, header, index, section):
        if header not in self._model:
            self._model[header] = list()
        self._model[header].insert(index, section)



import unittest
from collections import OrderedDict
from decimal import Decimal
from jesth import box, error
from jesth.section import Section
from jesth.converter import ValueConverter


TEXT_BODY = """\
# welcome
project = 'Jesth'
author = 'alex rustic'
pi = 3.141_592_653_589_793_238_46E0\
"""

DICT_BODY = OrderedDict([(box.CommentID(), box.Comment("# welcome")),
                         ("project", box.RawString("Jesth")),
                         ("author", box.RawString("alex rustic")),
                         ("pi", Decimal("3.141_592_653_589_793_238_46E0"))])


BASE_BODY = ["# welcome",
             "project = 'Jesth'",
             "author = 'alex rustic'",
             "pi = 3.141_592_653_589_793_238_46E0"]


class TestSectionClass(unittest.TestCase):

    def test_default_init(self):
        section = Section("header")
        self.assertEqual("header", section.header)
        self.assertEqual(list(), section.body)
        self.assertIsNotNone(section.value_converter)

    def test_init_with_dict_body(self):
        section = Section("header", DICT_BODY)
        r = section.body
        expected = BASE_BODY
        self.assertEqual(expected, r)

    def test_init_with_text_body(self):
        section = Section("header", TEXT_BODY)
        r = section.body
        expected = BASE_BODY
        self.assertEqual(expected, r)

    def test_init_with_base_body(self):
        section = Section("header", BASE_BODY)
        r = section.body
        expected = BASE_BODY
        self.assertEqual(expected, r)

    def test_init_with_custom_value_converter(self):
        value_converter = ValueConverter(float_decoder=Decimal)
        section = Section("", value_converter=value_converter)
        self.assertIs(Decimal, section.value_converter.float_decoder)

    def test_update_body(self):
        section = Section("header")
        invalid_base_body = 3.14
        with self.assertRaises(error.ConversionError):
            # body should be a list of string, a dict, or a string
            section.update(invalid_base_body)

    def test_make_dict_body(self):
        # test with empty body
        section = Section("")
        r = section.make_dict()
        self.assertEqual(OrderedDict(), r)
        # empty body + default parameter set
        default = {"is_empty": True}
        r = section.make_dict(default)
        self.assertIs(default, r)
        # == test with existent and compatible body
        section.update(BASE_BODY)
        # compact_mode = true
        r = section.make_dict(strict=True)
        self.assertIsNotNone(r)
        self.assertFalse(_find_comment(r))
        # compact_mode = false
        r = section.make_dict(strict=False)
        self.assertIsNotNone(r)
        self.assertTrue(_find_comment(r))

    def test_make_dict_body_with_incompatible_body(self):
        section = Section("", ["hello world"])
        with self.assertRaises(Exception):
            section.make_dict()

    def test_get_dict_body(self):
        # test with empty body
        section = Section("")
        r = section.get_dict()
        self.assertEqual(OrderedDict(), r)
        # empty body + default parameter set
        default = {"is_empty": True}
        r = section.get_dict(default)
        self.assertIs(default, r)
        # == test with existent and compatible body
        section.update(BASE_BODY)
        # compact_mode = true
        r = section.get_dict(strict=True)
        self.assertIsNotNone(r)
        self.assertFalse(_find_comment(r))
        # compact_mode = false
        r = section.get_dict(strict=False)
        self.assertIsNotNone(r)
        self.assertTrue(_find_comment(r))

    def test_get_dict_body_with_incompatible_body(self):
        section = Section("", ["hello world"])
        fallback = {"something_wrong": True}
        r = section.get_dict(fallback=fallback)
        self.assertEqual(fallback, r)


def _find_comment(body, comment="# welcome"):
    for key, val in body.items():
        if val == comment:
            return True
    return False


if __name__ == '__main__':
    unittest.main()



import unittest
import tempfile
import pathlib
from decimal import Decimal
from jesth import parse, read, ValueConverter


RAW_DOCUMENT = """\
Top anonymous
section

[user]
id = 1
name = 'alex'

[info]
project = 'jesth'
repository = 'https://github.com/pyrustic/jesth'

[user]
id = 2
name = 'rustic'\
"""


USER_AND_INFO = """\
[user]
id = 1
name = 'alex'

[info]
project = 'jesth'
repository = 'https://github.com/pyrustic/jesth'

[user]
id = 2
name = 'rustic'\
"""


USER_ONLY = """\
[user]
id = 1
name = 'alex'


[user]
id = 2
name = 'rustic'\
"""


class TestDocumentCreationWithParseFunc(unittest.TestCase):

    def test_with_default_args(self):
        document = parse(RAW_DOCUMENT)
        expected_headers = {"", "user", "info"}
        self.assertEqual(expected_headers, document.headers)
        self.assertIsNotNone(document.value_converter)
        self.assertEqual(4, len(document.sections))

    def test_with_custom_value_converter(self):
        value_converter = ValueConverter(float_decoder=Decimal)
        document = parse(RAW_DOCUMENT, value_converter=value_converter)
        self.assertIs(Decimal, document.value_converter.float_decoder)


class TestFileBasedDocument(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.write(RAW_DOCUMENT.encode("utf-8"))
        file.close()
        self._path_str = file.name
        self._path_obj = pathlib.Path(self._path_str)

    def tearDown(self):
        self._path_obj.unlink()

    def test_document_creation_with_read_func(self):
        # create temporary file
        document = read(self._path_str)
        expected_headers = {"", "user", "info"}
        self.assertEqual(expected_headers, document.headers)
        self.assertIsNotNone(document.value_converter)
        self.assertEqual(self._path_str, document.path)
        # with custom ValueConverter
        value_converter = ValueConverter(float_decoder=Decimal)
        document = read(self._path_str, value_converter=value_converter)
        self.assertIs(Decimal, document.value_converter.float_decoder)

    def test_save_method(self):
        document = read(self._path_str)
        document.append("new_header")
        r = document.save()
        self.assertTrue(r)
        # check
        with open(self._path_str, "r") as file:
            data = file.read()
        expected = RAW_DOCUMENT + "\n\n[new_header]"
        self.assertEqual(expected, data)


class TestDocumentCloning(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.close()
        self._path_str = file.name
        self._path_obj = pathlib.Path(self._path_str)

    def tearDown(self):
        self._path_obj.unlink()

    def test_save_to_method(self):
        document = parse(RAW_DOCUMENT)
        document.save_to(self._path_str)
        # check
        with open(self._path_str, "r") as file:
            data = file.read()
        self.assertEqual(data, RAW_DOCUMENT)


class TestDocumentRenderMethod(unittest.TestCase):

    def test_render_all_doc(self):
        document = parse(RAW_DOCUMENT)
        # render all doc
        r = document.render()
        expected = RAW_DOCUMENT
        self.assertEqual(expected, r)

    def test_render_specific_section(self):
        document = parse(RAW_DOCUMENT)
        # render specific section
        r = document.render("user", spacing=2)
        expected = USER_ONLY
        self.assertEqual(expected, r)

    def test_render_selected_sections(self):
        document = parse(RAW_DOCUMENT)
        # render specific sections
        r = document.render("user", "info", spacing=1)
        expected = USER_AND_INFO
        self.assertEqual(expected, r)
        # render specific sections
        r = document.render("", "info", "user")
        expected = RAW_DOCUMENT
        self.assertEqual(expected, r)


class TestDocumentDataInsertionMethods(unittest.TestCase):

    def test_insert_at_begin(self):
        text = "[section1]\n\n[section2]"
        # insert at 0
        document = parse(text)
        document.insert(0, "new_section")
        expected = "[new_section]\n\n" + text
        self.assertEqual(expected, document.render())

    def test_insert_at_end(self):
        text = "[section1]\n\n[section2]"
        # insert at the end
        document = parse(text)
        document.insert(-1, "new_section")
        expected = text + "\n\n[new_section]"
        self.assertEqual(expected, document.render())

    def test_insert_at_middle(self):
        text = "[section1]\n\n[section2]"
        # insert at the middle
        document = parse(text)
        document.insert(1, "new_section")
        expected = "[section1]\n\n[new_section]\n\n[section2]"
        self.assertEqual(expected, document.render())

    def test_append_method(self):
        document = parse(USER_ONLY)
        document.append("new_section")
        expected = USER_ONLY + "\n\n\n[new_section]"
        self.assertEqual(expected, document.render(spacing=2))


class TestDocumentDataRetrivalMethods(unittest.TestCase):

    def test_get_method(self):
        # default
        document = parse(USER_ONLY)
        user = document.get("user")
        expected_body = ["id = 2", "name = 'rustic'"]
        self.assertEqual(expected_body, user.body)
        # specific index
        user = document.get("user", 0)
        expected_body = ["id = 1", "name = 'alex'"]
        self.assertEqual(expected_body, user.body)

    def test_get_all_method(self):
        document = parse(USER_ONLY)
        users = document.get_all("user")
        expected_bodies = (["id = 1", "name = 'alex'"],
                           ["id = 2", "name = 'rustic'"])
        for i, user in enumerate(users):
            self.assertEqual(expected_bodies[i], user.body)

    def test_count_method(self):
        document = parse(USER_ONLY)
        self.assertEqual(2, document.count("user"))


class TestDocumentDataSuppressionMethods(unittest.TestCase):

    def test_remove_section_with_default_args(self):
        document = parse(USER_ONLY)
        self.assertEqual(2, document.count("user"))
        document.remove("user")
        user = document.get("user")
        expected_body = ["id = 1", "name = 'alex'"]
        self.assertEqual(expected_body, user.body)
        self.assertEqual(1, document.count("user"))

    def test_remove_section_at_relative_index(self):
        document = parse(USER_ONLY)
        self.assertEqual(2, document.count("user"))
        document.remove("user", 0)
        user = document.get("user")
        expected_body = ["id = 2", "name = 'rustic'"]
        self.assertEqual(expected_body, user.body)
        self.assertEqual(1, document.count("user"))

    def test_remove_all_method(self):
        document = parse(USER_AND_INFO)
        self.assertEqual(2, document.count("user"))
        self.assertEqual(1, document.count("info"))
        document.remove_all("user")
        self.assertEqual(0, document.count("user"))
        self.assertEqual(1, document.count("info"))


if __name__ == '__main__':
    unittest.main()



import unittest
import jesth


class TestRootImports(unittest.TestCase):

    def test_wrong_import(self):
        with self.assertRaises(AttributeError):
            jesth.abracadabra

    def test(self):
        jesth.Document
        jesth.Section
        jesth.parse
        jesth.render
        jesth.read
        jesth.write
        jesth.make_dict
        jesth.flatten_dict
        jesth.ValueConverter
        jesth.split_key_value
        jesth.Parser
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()



"""Private miscellaneous functions. Only "split_key_value" is public"""
import os
import os.path
from collections import namedtuple
from jesth import error, const


def split_key_value(line, sep="=", strip_whitespace=True):
    """
    Split a line (string) into key and value parts.
    The string must have a separator (defaults to "=").
    An example of a valid string is "name = John Doe ".
    The return will be: ("name": "John Doe")

    [parameters]
    - line: the string to split
    - sep: the separator, by default it is "="
    - strip_whitespace: boolean to tell whether you want whitespace at each
    side of the key and value to be stripped out or not.
    By default, the value is True.

    [return]
    Returns a ("key", "value") tuple if everything is Ok

    [exceptions]
    - error.Error: if the separator, the key, or the value is missing,
    """
    cache = line.split(sep, maxsplit=1)
    if len(cache) != 2:
        msg = "Missing separator character '{}' in '{}'"
        msg = msg.format(sep, line)
        raise error.Error(msg)
    key, value = cache
    if not key or key.isspace():
        msg = "Missing key in '{}'".format(line)
        raise error.Error(msg)
    if not value or value.isspace():
        msg = "Missing value in '{}'".format(line)
        raise error.Error(msg)
    if strip_whitespace:
        key, value = key.strip(), value.strip()
    return key, value


def ensure_parent_dir(path):
    """Make sure that parent dir exists (create it if isn't yet created)"""
    parent = os.path.dirname(path)
    try:
        os.makedirs(parent)
    except FileExistsError as e:
        pass


def update_cached_refs(value, cached_refs):
    if value in cached_refs:
        raise error.Error("Circular reference isn't allowed !")
    else:
        cached_refs.append(value)


def count_indents(line):
    i = 0
    for char in line:
        if char == " ":
            i += 1
            continue
        break
    x = i / const.INDENT_WIDTH
    y = int(x)
    if x > y:
        raise error.IndentError
    return y


def clean_leading_backslashes(line):
    """
    Remove one leading backslash only when the immediately-or-lately first non-backlash
    character on the line is an opening square bracket.
    """
    if not line or not line.startswith("\\"):
        return line
    if line.startswith(r"\["):
        line = line.replace(r"\[", "[", 1)
        return line
    char = None
    for i, char in enumerate(line):
        if char != "\\":
            break
    if char != "[":
        return line
    return line[1:]


def add_leading_backslashes(line):
    """
    Add one leading backslash only when the immediately-or-lately first non-backlash
    character on the line is an opening square bracket.
    """
    if not line:
        return line
    is_candidate = False
    for char in line:
        if char == "[":
            is_candidate = True
            break
        if char != "\\":
            break
    return "\\" + line if is_candidate else line


def get_headers(sections):
    "returns the set of headers from this list of sections"
    headers = set()
    for section in sections:
        headers.add(section.header)
    return headers


def decode_unicode(text):
    """Take in input some string that might have Unicode escape sequences.
    Output the same string with unicode escape sequences converted into
    the actual characters that they represent"""
    return text.encode("latin-1", "backslashreplace").decode("unicode-escape")


def encode_unicode(text, codec="latin-1"):
    """Convert a string into another where non-latin characters are
    replaced with Unicode escape sequences"""
    return text.encode(codec, "backslashreplace").decode(codec)


def tidy_up_float(s, width=3):
    """Tidy up a float number (str or float or decimal.Decimal)
    Example: 3.141234 -> 3.141_234
    """
    s = s if isinstance(s, str) else str(s)
    # make sure that s isn't "nan", "inf", "-inf", "infinity", ...
    for char in s:
        if char not in ("-01234.56789+eE"):
            return s
    info = parse_float(s)
    tidy_left_mantissa = tidy_up_int(info.left_mantissa)
    tidy_right_mantissa = ""
    if info.right_mantissa:
        cache = _tidy_up_right_mantissa(info.right_mantissa, width)
        tidy_right_mantissa = "." + cache
    tidy_exponent = "E" + info.exponent if info.exponent else ""
    result = tidy_left_mantissa + tidy_right_mantissa + tidy_exponent
    return result


def _tidy_up_right_mantissa(s, width):
    if not s:
        return ""
    cache = list()
    group = list()
    i = 0
    for char in s:
        if i == width:
            cache.append("".join(group))
            group = list()
            i = 0
        group.append(char)
        i += 1
    if group:
        cache.append("".join(group))
    return "_".join(cache)


def tidy_up_int(s, width=3):
    """Tidy up some int number (str or int)
    Example: 300141234 -> 300_141_234
    """
    s = s if isinstance(s, str) else str(s)
    tidy = list()
    is_negative = True if s[0] == "-" else False
    s = s.lstrip("+-")
    group = list()
    i = 0
    for char in reversed(s):
        if i == width:
            tidy.insert(0, "".join(group))
            group = list()
            i = 0
        group.insert(0, char)
        i += 1
    if group:
        tidy.insert(0, "".join(group))
    tidy = "_".join(tidy)
    result = "-" + tidy if is_negative else tidy
    return result


def tidy_up_hex(s, width=4):
    """Tidy up some hex int number (str or int)
    Example: 0xA3001F12A34 -> 0xA30_01F1_2A34
    """
    s = s if isinstance(s, str) else str(s)
    is_negative = True if s[0] == "-" else False
    s = s.lstrip("+-")
    s = s[2:]
    s = tidy_up_int(s, width=width)
    s = "0x" + s.upper()
    s = "-" + s if is_negative else s
    return s


def tidy_up_bin(s, width=4):
    """Tidy up some bin int number (str or int)
    Example: 0b10011101011 -> 0b100_1110_1011
    """
    s = s if isinstance(s, str) else str(s)
    is_negative = True if s[0] == "-" else False
    s = s.lstrip("+-")
    s = s[2:]
    s = tidy_up_int(s, width=width)
    s = "0b" + s
    s = "-" + s if is_negative else s
    return s


def tidy_up_oct(s, width=4):
    """Tidy up some oct int number (str or int)
    Example: 0o45631456202 -> 0o456_3145_6202
    """
    s = s if isinstance(s, str) else str(s)
    is_negative = True if s[0] == "-" else False
    s = s.lstrip("+-")
    s = s[2:]
    s = tidy_up_int(s, width=width)
    s = "0o" + s
    s = "-" + s if is_negative else s
    return s


FloatParts = namedtuple('FloatParts', ['left_mantissa', 'right_mantissa', 'exponent'])


def parse_float(s):
    """Parse a float number (string or decimal.Decimal or float), returns
    an instance of this namedtuple:
    FloatParts = namedtuple('FloatParts', ['left_mantissa', 'right_mantissa', 'exponent'])"""
    s = _prepare_float(s)
    # Split s at the exponent separator
    parts = s.split("E", 1)
    mantissa = parts[0]
    exponent = parts[1] if len(parts) == 2 else ""
    exponent = (exponent[1:] if exponent and exponent[0] == "+" 
                else exponent)
    # Split mantissa at the dot separator
    left_mantissa, right_mantissa = _parse_mantissa(mantissa)
    # Return the named tuple    
    return FloatParts(left_mantissa=left_mantissa, right_mantissa=right_mantissa, exponent=exponent)


def _prepare_float(s):
    if not isinstance(s, str):
        s = str(s)
    # Remove underscores, whitespaces, upper e
    cache = list()
    for c in s:
        if c in ("_", " "):
            continue
        if c == "e":
            c = "E"
        cache.append(c)
    return "".join(cache)


def _parse_mantissa(s):
    parts = s.split(".", 1)
    left_mantissa = parts[0]
    right_mantissa = parts[1] if len(parts) == 2 else ""
    return left_mantissa, right_mantissa

